/*
 * Copyright 2011 LMAX Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.lmax.disruptor.wizard;

import com.lmax.disruptor.AbstractEvent;
import com.lmax.disruptor.EventHandler;
import com.lmax.disruptor.EventProcessor;
import com.lmax.disruptor.DependencyBarrier;

import java.util.*;

class EventProcessorRepository<T extends AbstractEvent> implements Iterable<EventProcessorInfo<T>>
{
    private final Map<EventHandler, EventProcessorInfo<T>> eventProcessorInfoByHandler = new IdentityHashMap<EventHandler, EventProcessorInfo<T>>();
    private final Map<EventProcessor, EventProcessorInfo<T>> eventProcessorInfoByEventProcessor = new IdentityHashMap<EventProcessor, EventProcessorInfo<T>>();

    public void add(EventProcessor eventprocessor, EventHandler<T> handler, final DependencyBarrier barrier)
    {
        final EventProcessorInfo<T> eventProcessorInfo = new EventProcessorInfo<T>(eventprocessor, handler, barrier);
        eventProcessorInfoByHandler.put(handler, eventProcessorInfo);
        eventProcessorInfoByEventProcessor.put(eventprocessor, eventProcessorInfo);
    }

    public EventProcessor[] getLastEventProcessorsInChain()
    {
        List<EventProcessor> lastEventProcessors = new ArrayList<EventProcessor>();
        for (EventProcessorInfo<T> eventProcessorInfo : eventProcessorInfoByHandler.values())
        {
            if (eventProcessorInfo.isEndOfChain())
            {
                lastEventProcessors.add(eventProcessorInfo.getEventProcessor());
            }
        }
        return lastEventProcessors.toArray(new EventProcessor[lastEventProcessors.size()]);
    }

    public EventProcessor getEventProcessorFor(final EventHandler<T> handler)
    {
        final EventProcessorInfo eventprocessorInfo = getEventProcessorInfo(handler);
        return eventprocessorInfo != null ? eventprocessorInfo.getEventProcessor() : null;
    }

    public void unmarkEventProcessorsAsEndOfChain(final EventProcessor... barrierEventProcessors)
    {
        for (EventProcessor barrierEventProcessor : barrierEventProcessors)
        {
            getEventProcessorInfo(barrierEventProcessor).usedInBarrier();
        }
    }

    public Iterator<EventProcessorInfo<T>> iterator()
    {
        return eventProcessorInfoByHandler.values().iterator();
    }

    public DependencyBarrier getBarrierFor(final EventHandler<T> handler)
    {
        final EventProcessorInfo<T> eventProcessorInfo = getEventProcessorInfo(handler);
        return eventProcessorInfo != null ? eventProcessorInfo.getBarrier() : null;
    }

    private EventProcessorInfo<T> getEventProcessorInfo(final EventHandler<T> handler)
    {
        return eventProcessorInfoByHandler.get(handler);
    }

    private EventProcessorInfo<T> getEventProcessorInfo(final EventProcessor barrierEventProcessor)
    {
        return eventProcessorInfoByEventProcessor.get(barrierEventProcessor);
    }
}
